/*
 * linux/arch/arm/drivers/scsi/powertec.c
 *
 * Copyright (C) 1997 Russell King
 *
 * This driver is based on experimentation.  Hence, it may have made
 * assumptions about the particular card that I have available, and
 * may not be reliable!
 *
 * Changelog:
 *  01-10-1997	RMK	Created, READONLY version
 */

#include <linux/module.h>
#include <linux/blk.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/unistd.h>
#include <linux/stat.h>

#include <asm/delay.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/ecard.h>

#include "sd.h"
#include "hosts.h"
#include "powertec.h"

/* Hmm - this should go somewhere else */
#define BUS_ADDR(x) ((((unsigned long)(x)) << 2) + IO_BASE)

/* Configuration */
#define XTALFREQ		40
#define INT_POLARITY		CTRL_INT_HIGH

/*
 * List of devices that the driver will recognise
 */
#define POWERTECSCSI_LIST	{ MANU_ALSYSTEMS, PROD_ALSYS_SCSIATAPI }
#define POWERTEC_FAS216_OFFSET	0

/*
 * Version
 */
#define VER_MAJOR	0
#define VER_MINOR	0
#define VER_PATCH	0

static struct expansion_card *ecs[MAX_ECARDS];

static struct proc_dir_entry proc_scsi_powertec = {
	PROC_SCSI_QLOGICISP, 6, "powertec",
	S_IFDIR | S_IRUGO | S_IXUGO, 2
};

/*
 * Function: void powertecscsi_intr (int irq, void *dev_id, struct pt_regs *regs)
 * Purpose : handle interrupts from Powertec SCSI card
 * Params  : irq - interrupt number
 *	     dev_id - user-defined (Scsi_Host structure)
 *	     regs - processor registers at interrupt
 */
static void powertecscsi_intr (int irq, void *dev_id, struct pt_regs *regs)
{
    struct Scsi_Host *instance = (struct Scsi_Host *)dev_id;

    fas216_intr (instance);
}

/*
 * Function: int powertecscsi_dma_setup (instance, SCpnt, direction)
 * Purpose : initialises DMA/PIO
 * Params  : instance  - host
 *	     SCpnt     - command
 *	     direction - DMA on to/off of card
 * Returns : 0 if we should not set CMD_WITHDMA for transfer info command
 */
static int powertecscsi_dma_setup (struct Scsi_Host *instance, Scsi_Pointer *SCp, dmadir_t direction)
{
    /*
     * We don't do DMA, we only do PSEUDO dma
     */
    return 0;
}

/*
 * Function: int powertecscsi_dma_pseudo (instance, SCpnt, direction, transfer)
 * Purpose : handles pseudo DMA
 * Params  : instance  - host
 *	     SCpnt     - command
 *	     direction - DMA on to/off of card
 *	     transfer  - minimum number of bytes we expect to transfer
 * Returns : bytes transfered
 */
static int
powertecscsi_dma_pseudo (struct Scsi_Host *instance, Scsi_Pointer *SCp,
			 dmadir_t direction, int transfer)
{
    PowerTecScsi_Info *info = (PowerTecScsi_Info *)instance->hostdata;
    unsigned int length;
    unsigned char *addr;

    length = SCp->this_residual;
    addr = SCp->ptr;

    if (direction == DMA_OUT) {
    } else {
    }

    return SCp->this_residual - length;
}

/*
 * Function: int powertecscsi_dma_stop (instance, SCpnt)
 * Purpose : stops DMA/PIO
 * Params  : instance  - host
 *	     SCpnt     - command
 */
static void powertecscsi_dma_stop (struct Scsi_Host *instance, Scsi_Pointer *SCp)
{
    /*
     * no DMA to stop
     */
}

/*
 * Function: int powertecscsi_detect (Scsi_Host_Template * tpnt)
 * Purpose : initialises PowerTec SCSI driver
 * Params  : tpnt - template for this SCSI adapter
 * Returns : >0 if host found, 0 otherwise.
 */
int powertecscsi_detect (Scsi_Host_Template *tpnt)
{
    static const card_ids powertecscsi_cids[] = { POWERTECSCSI_LIST, { 0xffff, 0xffff} };
    int count = 0;
    struct Scsi_Host *instance;
  
    tpnt->proc_dir = &proc_scsi_powertec;
    memset (ecs, 0, sizeof (ecs));

    ecard_startfind ();

    while (1) {
    	PowerTecScsi_Info *info;

	ecs[count] = ecard_find (0, powertecscsi_cids);
	if (!ecs[count])
	    break;

	ecard_claim (ecs[count]);

	instance = scsi_register (tpnt, sizeof (PowerTecScsi_Info));
	if (!instance) {
	    ecard_release (ecs[count]);
	    break;
	}

	instance->io_port = ecard_address (ecs[count], ECARD_MEMC, 0);
	instance->irq = ecs[count]->irq;

#ifdef required
	ecs[count]->irqaddr = (unsigned char *)BUS_ADDR(instance->io_port);
	ecs[count]->irqmask = !IRQ mask!;
#endif

	request_region (instance->io_port + POWERTEC_FAS216_OFFSET,
			16, "powertec2-fas");

	if (request_irq (instance->irq, powertecscsi_intr, SA_INTERRUPT, "powertec2", instance)) {
	    printk ("scsi%d: IRQ%d not free, interrupts disabled\n",
		      instance->host_no, instance->irq);
	}

	info = (PowerTecScsi_Info *)instance->hostdata;
	info->info.scsi.io_port	= instance->io_port + POWERTEC_FAS216_OFFSET;
	info->info.scsi.irq	= instance->irq;
	info->info.ifcfg.clockrate = XTALFREQ;
	info->info.ifcfg.select_timeout = 255;
	info->info.dma.setup	= powertecscsi_dma_setup;
	info->info.dma.pseudo	= powertecscsi_dma_pseudo;
	info->info.dma.stop	= powertecscsi_dma_stop;

	fas216_init (instance);
	++count;
    }
    return count;
}

/*
 * Function: int powertecscsi_release (struct Scsi_Host * host)
 * Purpose : releases all resources used by this adapter
 * Params  : host - driver host structure to return info for.
 * Returns : nothing
 */
int powertecscsi_release (struct Scsi_Host *instance)
{
    int i;

    fas216_release (instance);

    if (instance->irq != 255)
	free_irq (instance->irq, instance);
    release_region (instance->io_port + POWERTEC_FAS216_OFFSET, 16);

    for (i = 0; i < MAX_ECARDS; i++)
	if (ecs[i] && instance->io_port == ecard_address (ecs[i], ECARD_MEMC, 0))
		ecard_release (ecs[i]);
    return 0;
}

/*
 * Function: const char *powertecscsi_info (struct Scsi_Host * host)
 * Purpose : returns a descriptive string about this interface,
 * Params  : host - driver host structure to return info for.
 * Returns : pointer to a static buffer containing null terminated string.
 */
const char *powertecscsi_info (struct Scsi_Host *host)
{
    PowerTecScsi_Info *info = (PowerTecScsi_Info *)host->hostdata;
    static char string[100], *p;

    p = string;
    p += sprintf (string, "%s at port %X irq %d v%d.%d.%d scsi %s",
		  host->hostt->name, host->io_port, host->irq,
		  VER_MAJOR, VER_MINOR, VER_PATCH,
		  info->info.scsi.type);

    return string;
}

/*
 * Function: int powertecscsi_proc_info (char *buffer, char **start, off_t offset,
 *					 int length, int host_no, int inout)
 * Purpose : Return information about the driver to a user process accessing
 *	     the /proc filesystem.
 * Params  : buffer - a buffer to write information to
 *	     start  - a pointer into this buffer set by this routine to the start
 *		      of the required information.
 *	     offset - offset into information that we have read upto.
 *	     length - length of buffer
 *	     host_no - host number to return information for
 *	     inout  - 0 for reading, 1 for writing.
 * Returns : length of data written to buffer.
 */
int powertecscsi_proc_info (char *buffer, char **start, off_t offset,
			    int length, int host_no, int inout)
{
    int pos, begin, first;
    struct Scsi_Host *host = scsi_hostlist;
    PowerTecScsi_Info *info;
    Scsi_Device *scd;

    while (host) {
	if (host->host_no == host_no)
	    break;
	host = host->next;
    }
    if (!host)
	return 0;

    info = (PowerTecScsi_Info *)host->hostdata;
    if (inout == 1)
      return -EINVAL;

    begin = 0;
    pos = sprintf (buffer,
			"PowerTec SCSI driver version %d.%d.%d\n",
			VER_MAJOR, VER_MINOR, VER_PATCH);
    pos += sprintf (buffer + pos,
			"Address: %08X          IRQ : %d\n"
			"FAS    : %s\n\n"
			"Statistics:\n",
			host->io_port, host->irq, info->info.scsi.type);

    pos += sprintf (buffer+pos,
			"Queued commands: %-10ld   Issued commands: %-10ld\n"
			"Done commands  : %-10ld   Reads          : %-10ld\n"
			"Writes         : %-10ld   Others         : %-10ld\n"
			"Disconnects    : %-10ld   Aborts         : %-10ld\n"
			"Resets         : %-10ld\n",
			 info->info.stats.queues,      info->info.stats.removes,
			 info->info.stats.fins,        info->info.stats.reads,
			 info->info.stats.writes,      info->info.stats.miscs,
			 info->info.stats.disconnects, info->info.stats.aborts,
			 info->info.stats.resets);

    first = 1;
    for (scd = scsi_devices; scd; scd = scd->next) {
	if (scd->host == host) {
	    int len;

	    if (first) {
		pos += sprintf (buffer+pos, "\nAttached devices:\n");
		first = 0;
	    }

	    proc_print_scsidevice (scd, buffer, &len, pos);
	    pos += len;
	    pos += sprintf (buffer+pos, "Extensions: ");
	    if (scd->tagged_supported)
		pos += sprintf (buffer+pos, "TAG %sabled [%d] ",
				scd->tagged_queue ? "en" : "dis",
				scd->current_tag);
	    pos += sprintf (buffer+pos, "\n");

	    if (pos + begin < offset) {
		begin += pos;
		pos = 0;
	    }
	    if (pos + begin > offset + length)
		break;
	}
    }

    *start = buffer + (offset - begin);
    pos -= offset - begin;
    if (pos > length)
	pos = length;

    return pos;
}

#ifdef MODULE
Scsi_Host_Template driver_template = POWERTECSCSI;

#include "scsi_module.c"
#endif
